# -*- coding: utf-8 -*-
# vim: set noai syntax=python ts=4 sw=4:
#
# Copyright (c) 2018-2023 Linh Pham
# wwdtm is released under the terms of the Apache License 2.0
"""Wait Wait Don't Tell Me! Stats Show Data Retrieval Functions
"""
import datetime
import dateutil.parser as date_parser
from decimal import Decimal
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple, Union

from mysql.connector import connect, DatabaseError
from wwdtm.show.info import ShowInfo
from wwdtm.show.info_multiple import ShowInfoMultiple
from wwdtm.show.utility import ShowUtility
from wwdtm.validation import valid_int_id


class Show:
    """This class contains functions used to retrieve show data from a
    copy of the Wait Wait Stats database.

    :param connect_dict: Dictionary containing database connection
        settings as required by mysql.connector.connect
    :param database_connection: mysql.connector.connect database
        connection
    """

    def __init__(
        self,
        connect_dict: Optional[Dict[str, Any]] = None,
        database_connection: Optional[connect] = None,
    ):
        """Class initialization method."""
        if connect_dict:
            self.connect_dict = connect_dict
            self.database_connection = connect(**connect_dict)
        elif database_connection:
            if not database_connection.is_connected():
                database_connection.reconnect()

            self.database_connection = database_connection

        try:
            query = (
                "SHOW COLUMNS FROM ww_showpnlmap WHERE Field = 'panelistscore_decimal';"
            )
            cursor = self.database_connection.cursor()
            cursor.execute(query)
            result = cursor.fetchone()
            cursor.close()
            if result:
                self.panelist_decimal_column: bool = True
            else:
                self.panelist_decimal_column: bool = False
        except DatabaseError:
            self.panelist_decimal_column: bool = False

        self.info = ShowInfo(database_connection=self.database_connection)
        self.info_multiple = ShowInfoMultiple(
            database_connection=self.database_connection
        )
        self.utility = ShowUtility(database_connection=self.database_connection)

    def retrieve_all(self) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show ID,
        show date, Best Of and Repeat Show information for all shows.

        :return: List of all shows and their corresponding information.
            If show information could not be retrieved, an empty list
            will be returned.
        """
        query = """
            SELECT showid AS id, showdate AS date,
            bestof AS best_of, repeatshowid AS repeat_show_id
            FROM ww_shows
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query)
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        shows = []
        for row in results:
            show = {
                "id": row.id,
                "date": row.date.isoformat(),
                "best_of": bool(row.best_of),
                "repeat_show": bool(row.repeat_show_id),
            }

            if row.repeat_show_id:
                show["original_show_id"] = row.repeat_show_id
                show["original_show_date"] = self.utility.convert_id_to_date(
                    row.repeat_show_id
                )

            shows.append(show)

        return shows

    def retrieve_all_details(
        self, include_decimal_scores: bool = False
    ) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information
        for all shows.

        :param include_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: List of all shows and their corresponding details.
            If show information could not be retrieved, an empty list
            will be returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return []

        info = self.info_multiple.retrieve_core_info_all()

        if not info:
            return []

        panelists = self.info_multiple.retrieve_panelist_info_all(
            include_decimal_scores=include_decimal_scores
        )
        bluffs = self.info_multiple.retrieve_bluff_info_all()
        guests = self.info_multiple.retrieve_guest_info_all()

        shows = []
        for show in info:
            if info[show]["id"] in panelists:
                info[show]["panelists"] = panelists[info[show]["id"]]
            else:
                info[show]["panelists"] = []

            if info[show]["id"] in bluffs:
                info[show]["bluff"] = bluffs[info[show]["id"]]
            else:
                info[show]["bluff"] = []

            if info[show]["id"] in guests:
                info[show]["guests"] = guests[info[show]["id"]]
            else:
                info[show]["guests"] = []

            shows.append(info[show])

        return shows

    def retrieve_all_ids(self) -> List[int]:
        """Returns a list of all show IDs from the database, sorted by
        show date.

        :return: List of all show IDs. If show IDs could not be
            retrieved, an empty list will be returned.
        """
        query = """
            SELECT showid FROM ww_shows ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query)
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [v[0] for v in results]

    def retrieve_all_dates(self) -> List[str]:
        """Returns a list of all show dates from the database, sorted
        by show date.

        :return: List of all show date strings. If show dates could not
            be retrieved, an empty list will be returned.
        """
        query = """
            SELECT showdate FROM ww_shows ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query)
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [v[0].isoformat() for v in results]

    def retrieve_all_dates_tuple(self) -> List[Tuple[int, int, int]]:
        """Returns a list of all show dates as a tuple of year, month,
        and day, sorted by show date.

        :return: List of allow show dates as a tuple of year, month
            and day. If show dates could not be retrieved, an empty list
            will be returned.
        """
        query = """
            SELECT YEAR(showdate), MONTH(showdate), DAY(showdate)
            FROM ww_shows
            ORDER BY YEAR(showdate) ASC, MONTH(showdate) ASC,
            DAY(showdate) ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query)
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [tuple(v) for v in results]

    def retrieve_all_show_years_months(self) -> List[str]:
        """Returns a list of all show years and months as a string,
        sorted by year and month.

        :return: List of all show years and month in ``YYYY-MM`` format.
            If show dates could not be retrieved, an empty list will be
            returned.
        """
        query = """
            SELECT DISTINCT YEAR(showdate), MONTH(showdate)
            FROM ww_shows
            ORDER BY YEAR(showdate) ASC, MONTH(showdate) ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query)
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [f"{v[0]}-{v[1]}" for v in results]

    def retrieve_all_shows_years_months_tuple(self) -> List[Tuple[int, int]]:
        """Returns a list of all show years and months as a tuple of
        year and month, sorted by year and month.

        :return: List of allow show dates as a tuple of year and month.
            If show dates could not be retrieved, an empty list will be
            returned.
        """
        query = """
            SELECT DISTINCT YEAR(showdate), MONTH(showdate)
            FROM ww_shows
            ORDER BY YEAR(showdate) ASC, MONTH(showdate) ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query)
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [tuple(v) for v in results]

    @lru_cache(typed=True)
    def retrieve_by_date(self, year: int, month: int, day: int) -> Dict[str, Any]:
        """Returns a dictionary object containing show ID, show date,
        Best Of and Repeat Show information for the requested show date.

        :param year: Four-digit year
        :param month: One or two-digit month
        :param day: One or two-digit day
        :return: Dictionary containing show information. If show
            information could not be retrieved, an empty dictionary will
            be returned.
        """
        id_ = self.utility.convert_date_to_id(year, month, day)
        if not id_:
            return {}

        return self.retrieve_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_by_date_string(self, date_string: str) -> Dict[str, Any]:
        """Returns a dictionary object containing show ID, show date,
        Best Of and Repeat Show information for the requested show date
        string.

        :param date_string: Show date in ``YYYY-MM-DD`` format
        :return: Dictionary containing show information. If show
            information could not be retrieved, an empty dictionary will
            be returned.
        """
        try:
            parsed_date_string = date_parser.parse(date_string)
        except ValueError:
            return {}

        id_ = self.utility.convert_date_to_id(
            parsed_date_string.year, parsed_date_string.month, parsed_date_string.day
        )

        return self.retrieve_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_by_id(self, show_id: int) -> Dict[str, Any]:
        """Returns a dictionary object containing show ID, show date,
        Best Of and Repeat Show information for the requested show ID.

        :param show_id: Show ID
        :return: Dictionary containing show information. If show
            information could not be retrieved, an empty dictionary will
            be returned.
        """
        if not valid_int_id(show_id):
            return {}

        query = """
            SELECT showid AS id, showdate AS date,
            bestof AS best_of, repeatshowid AS repeat_show_id
            FROM ww_shows
            WHERE showid = %s
            LIMIT 1;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query, (show_id,))
        result = cursor.fetchone()
        cursor.close()

        if not result:
            return {}

        info = {
            "id": result.id,
            "date": result.date.isoformat(),
            "best_of": bool(result.best_of),
            "repeat_show": bool(result.repeat_show_id),
        }

        if result.repeat_show_id:
            info["original_show_id"] = result.repeat_show_id
            info["original_show_date"] = self.utility.convert_id_to_date(
                result.repeat_show_id
            )

        return info

    @lru_cache(typed=True)
    def retrieve_by_month_day(self, month: int, day: int) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing with show
        information for the requested month and day, sorted by year.

        :param month: One or two-digit month
        :param day: One or two-digit day
        :return: List of shows for the requested month, day, and
            corresponding information. If show information could not be
            retrieved, an empty list will be returned.
        """
        if not 1 <= month <= 12 or not 1 <= day <= 31:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE MONTH(showdate) = %s AND DAY(showdate) = %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(
            query,
            (
                month,
                day,
            ),
        )
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [self.retrieve_by_id(v[0]) for v in results]

    @lru_cache(typed=True)
    def retrieve_by_year(self, year: int) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing with show
        information for the requested year, sorted by show date.

        :param year: Four-digit year
        :return: List of shows for the requested year and corresponding
            information. If show information could not be retrieved,
            an empty list will be returned.
        """
        try:
            parsed_year = date_parser.parse(f"{year:04d}")
        except ValueError:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE YEAR(showdate) = %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query, (parsed_year.year,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [self.retrieve_by_id(v[0]) for v in results]

    @lru_cache(typed=True)
    def retrieve_by_year_month(self, year: int, month: int) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show
        information for the requested year and month, sorted by show
        date.

        :param year: Four-digit year
        :param month: One or two-digit month
        :return: List of shows for the requested year and month, and
            corresponding information. If show information could not be
            retrieved, a list of dictionaries will be returned.
        """
        try:
            parsed_year_month = date_parser.parse(f"{year:04d}-{month:02d}-01")
        except ValueError:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE YEAR(showdate) = %s AND MONTH(showdate) = %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(
            query,
            (
                parsed_year_month.year,
                parsed_year_month.month,
            ),
        )
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [self.retrieve_by_id(v[0]) for v in results]

    @lru_cache(typed=True)
    def retrieve_details_by_date(
        self, year: int, month: int, day: int, include_decimal_scores: bool = False
    ) -> Dict[str, Any]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information
        for the requested show date.

        :param year: Four-digit year
        :param month: One or two-digit month
        :param day: One or two digit day
        :return: Dictionary containing show information and details. If
            show information could not be retrieved, an empty dictionary
            will be returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return {}

        id_ = self.utility.convert_date_to_id(year, month, day)
        if not id_:
            return {}

        return self.retrieve_details_by_id(
            id_, include_decimal_scores=include_decimal_scores
        )

    @lru_cache(typed=True)
    def retrieve_details_by_date_string(
        self, date_string: str, include_decimal_scores: bool = False
    ) -> Dict[str, Any]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information
        for the requested show date string.

        :param date_string: Show date in ``YYYY-MM-DD`` format
        :param include_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: Dictionary containing show information and details. If
            show information could not be retrieved, an empty dictionary
            will be returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return {}

        try:
            parsed_date_string = date_parser.parse(date_string)
        except ValueError:
            return {}

        id_ = self.utility.convert_date_to_id(
            parsed_date_string.year, parsed_date_string.month, parsed_date_string.day
        )

        return self.retrieve_details_by_id(
            id_, include_decimal_scores=include_decimal_scores
        )

    @lru_cache(typed=True)
    def retrieve_details_by_id(
        self, show_id: int, include_decimal_scores: bool = False
    ) -> Dict[str, Any]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information
        for the requested show ID.

        :param show_id: Show ID
        :param include_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: Dictionary containing show information and details. If
            show information could not be retrieved, an empty dictionary
            will be returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return {}

        if not valid_int_id(show_id):
            return {}

        info = self.info.retrieve_core_info_by_id(show_id)
        if not info:
            return {}

        info["panelists"] = self.info.retrieve_panelist_info_by_id(
            show_id, include_decimal_scores=include_decimal_scores
        )
        info["bluff"] = self.info.retrieve_bluff_info_by_id(show_id)
        info["guests"] = self.info.retrieve_guest_info_by_id(show_id)

        return info

    @lru_cache(typed=True)
    def retrieve_details_by_month_day(
        self, month: int, day: int, include_decimal_scores: bool = False
    ) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information
        for the requested month and day, sorted by year.

        :param month: One or two-digit month
        :param day: One or two-digit day
        :param include_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: Dictionary containing show information and details. If
            show information could not be retrieved, an empty dictionary
            will be returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return []

        if not 1 <= month <= 12 or not 1 <= day <= 31:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE MONTH(showdate) = %s AND DAY(showdate) = %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(
            query,
            (
                month,
                day,
            ),
        )
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        show_ids = [v[0] for v in results]
        info = self.info_multiple.retrieve_core_info_by_ids(show_ids)

        if not info:
            return []

        shows = []
        for show in info:
            if info[show]:
                info[show]["panelists"] = self.info.retrieve_panelist_info_by_id(
                    info[show]["id"], include_decimal_scores=include_decimal_scores
                )
                info[show]["bluff"] = self.info.retrieve_bluff_info_by_id(
                    info[show]["id"]
                )
                info[show]["guests"] = self.info.retrieve_guest_info_by_id(
                    info[show]["id"]
                )
                shows.append(info[show])

        return shows

    @lru_cache(typed=True)
    def retrieve_details_by_year(
        self, year: int, include_decimal_scores: bool = False
    ) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information for
        the requested year, sorted by show date.

        :param year: Four-digit year
        :param include_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: List of shows for the requested year and corresponding
            details. If show information could not be retrieved, an
            empty list will be returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return []

        try:
            parsed_year = date_parser.parse(f"{year:04d}")
        except ValueError:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE YEAR(showdate) = %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query, (parsed_year.year,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        show_ids = [v[0] for v in results]
        info = self.info_multiple.retrieve_core_info_by_ids(show_ids)

        if not info:
            return []

        shows = []
        for show in info:
            if info[show]:
                info[show]["panelists"] = self.info.retrieve_panelist_info_by_id(
                    info[show]["id"], include_decimal_scores=include_decimal_scores
                )
                info[show]["bluff"] = self.info.retrieve_bluff_info_by_id(
                    info[show]["id"]
                )
                info[show]["guests"] = self.info.retrieve_guest_info_by_id(
                    info[show]["id"]
                )
                shows.append(info[show])

        return shows

    @lru_cache(typed=True)
    def retrieve_details_by_year_month(
        self, year: int, month: int, include_decimal_scores: bool = False
    ) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information for
        the requested year and month, sorted by show date.

        :param year: Four-digit year
        :param month: One or two-digit month
        :param include_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: List of shows for the requested year and month, and
            corresponding details. If show information could not be
            retrieved, an empty list will be returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return []

        try:
            parsed_year_month = date_parser.parse(f"{year:04d}-{month:02d}-01")
        except ValueError:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE YEAR(showdate) = %s AND MONTH(showdate) = %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(
            query,
            (
                parsed_year_month.year,
                parsed_year_month.month,
            ),
        )
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        show_ids = [v[0] for v in results]
        info = self.info_multiple.retrieve_core_info_by_ids(show_ids)

        if not info:
            return []

        shows = []
        for show in info:
            if info[show]:
                info[show]["panelists"] = self.info.retrieve_panelist_info_by_id(
                    info[show]["id"], include_decimal_scores=include_decimal_scores
                )
                info[show]["bluff"] = self.info.retrieve_bluff_info_by_id(
                    info[show]["id"]
                )
                info[show]["guests"] = self.info.retrieve_guest_info_by_id(
                    info[show]["id"]
                )
                shows.append(info[show])

        return shows

    @lru_cache(typed=True)
    def retrieve_months_by_year(self, year: int) -> List[int]:
        """Returns a list of show months available for the requested
        year, sorted by month.

        :param year: Four-digit year
        :return: List of available show months. If show information
            could not be retrieved, an empty list will be returned.
        """
        try:
            _ = date_parser.parse(f"{year:04d}")
        except ValueError:
            return []

        query = """
            SELECT DISTINCT MONTH(showdate)
            FROM ww_shows
            WHERE YEAR(showdate) = %s
            ORDER BY MONTH(showdate) ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query, (year,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [v[0] for v in results]

    @lru_cache(typed=True)
    def retrieve_recent(
        self,
        include_days_ahead: Optional[int] = 7,
        include_days_back: Optional[int] = 32,
    ) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show ID,
        show date, Best Of and Repeat Show information for recent shows.

        :param include_days_ahead: Number of days in the future to
            include, defaults to 7
        :param include_days_back: Number of days in the past to
            include, defaults to 32
        :return: List of recent shows and corresponding information. If
            show information could not be retrieved, an empty list will
            be returned.
        """
        try:
            past_days = int(include_days_back)
            future_days = int(include_days_ahead)
        except ValueError:
            return []

        try:
            past_date = datetime.datetime.now() - datetime.timedelta(days=past_days)
            future_date = datetime.datetime.now() + datetime.timedelta(days=future_days)
        except OverflowError:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE showdate >= %s AND showdate <= %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(
            query,
            (
                past_date.isoformat(),
                future_date.isoformat(),
            ),
        )
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [self.retrieve_by_id(v[0]) for v in results]

    @lru_cache(typed=True)
    def retrieve_recent_details(
        self,
        include_days_ahead: Optional[int] = 7,
        include_days_back: Optional[int] = 32,
        include_decimal_scores: bool = False,
    ) -> List[Dict[str, Any]]:
        """Returns a list of dictionary objects containing show ID,
        show date, host, scorekeeper, panelist and guest information
        for recent shows.

        :param include_days_ahead: Number of days in the future to
            include, defaults to 7
        :param include_days_back: Number of days in the past to
            include, defaults to 32
        :param include_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: List of recent shows and corresponding details. If show
            information could not be retrieved, an empty list will be
            returned.
        """
        if include_decimal_scores and not self.panelist_decimal_column:
            return []

        try:
            past_days = int(include_days_back)
            future_days = int(include_days_ahead)
        except ValueError:
            return []

        try:
            past_date = datetime.datetime.now() - datetime.timedelta(days=past_days)
            future_date = datetime.datetime.now() + datetime.timedelta(days=future_days)
        except OverflowError:
            return []

        query = """
            SELECT showid FROM ww_shows
            WHERE showdate >= %s AND showdate <= %s
            ORDER BY showdate ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(
            query,
            (
                past_date.isoformat(),
                future_date.isoformat(),
            ),
        )
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        show_ids = [v[0] for v in results]
        info = self.info_multiple.retrieve_core_info_by_ids(show_ids)

        if not info:
            return []

        shows = []
        for show in info:
            info[show]["panelists"] = self.info.retrieve_panelist_info_by_id(
                show, include_decimal_scores=include_decimal_scores
            )
            info[show]["bluff"] = self.info.retrieve_bluff_info_by_id(show)
            info[show]["guests"] = self.info.retrieve_guest_info_by_id(show)
            shows.append(info[show])

        return shows

    @lru_cache(typed=True)
    def retrieve_scores_by_year(
        self, year: int, use_decimal_scores: bool = False
    ) -> List[Tuple[str, Union[int, Decimal]]]:
        """Returns a list of tuples containing panelist scores for all
        shows in the requested year, sorted by show date.

        :param year: Four-digit year
        :param use_decimal_scores: Flag set to use panelist decimal scores,
            if available
        :return: List of tuples each containing show date and panelist
            scores. If show scores could not be retrieved, an empty list
            will be returned.
        """
        if use_decimal_scores and not self.panelist_decimal_column:
            return []

        try:
            _ = date_parser.parse(f"{year:04d}")
        except ValueError:
            return []

        if use_decimal_scores:
            query = """
                SELECT s.showdate AS date, pm.panelistscore_decimal AS score
                FROM ww_showpnlmap pm
                JOIN ww_shows s ON s.showid = pm.showid
                WHERE s.bestof = 0 AND s.repeatshowid IS NULL
                AND pm.panelistscore_decimal IS NOT NULL
                AND YEAR(s.showdate) = %s
                ORDER BY s.showdate ASC, pm.panelistscore_decimal ASC;
                """
        else:
            query = """
                SELECT s.showdate AS date, pm.panelistscore AS score
                FROM ww_showpnlmap pm
                JOIN ww_shows s ON s.showid = pm.showid
                WHERE s.bestof = 0 AND s.repeatshowid IS NULL
                AND pm.panelistscore IS NOT NULL
                AND YEAR(s.showdate) = %s
                ORDER BY s.showdate ASC, pm.panelistscore ASC;
                """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query, (year,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        shows = {}
        for row in results:
            date = row.date.isoformat()
            if date not in shows:
                shows[date] = []

            shows[date].append(row.score)

        show_scores = []
        for show in shows:
            show_score = shows[show]
            show_score.insert(0, show)
            show_scores.append(tuple(show_score))

        return show_scores

    def retrieve_years(self) -> List[int]:
        """Returns a list of available show years, sorted by year.

        :return: List of available show years. If show dates could not
            be retrieved, an empty list will be returned.
        """
        query = """
            SELECT DISTINCT YEAR(showdate)
            FROM ww_shows
            ORDER BY YEAR(showdate) ASC;
            """
        cursor = self.database_connection.cursor(dictionary=False)
        cursor.execute(query)
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        return [v[0] for v in results]



# -*- coding: utf-8 -*-
# vim: set noai syntax=python ts=4 sw=4:
#
# Copyright (c) 2018-2023 Linh Pham
# wwdtm is released under the terms of the Apache License 2.0
"""Type validation module"""


def valid_int_id(int_id: int) -> bool:
    """Validates an ID value as a signed 32-bit integer used
    in ID fields in MySQL tables.

    :param int_id: ID number to validate
    :return: True or False, based on if the integer falls inclusively
        between 0 and 2147483647
    """
    try:
        if not int_id:
            return False

        int_id_ = int(int_id)
    except ValueError:
        return False

    # Minimum value of a signed INT in MySQL/MariaDB is 0 and the
    # maximum value of signed INT type in MySQL/MariaDB is (2**31) - 1,
    # or 2147483647
    return 0 <= int_id_ <= (2**31 - 1)



# -*- coding: utf-8 -*-
# vim: set noai syntax=python ts=4 sw=4:
#
# Copyright (c) 2018-2023 Linh Pham
# wwdtm is released under the terms of the Apache License 2.0
"""Wait Wait Don't Tell Me! Stats Panelist Scores Retrieval Functions
"""
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple

from mysql.connector import connect
from wwdtm.panelist.utility import PanelistUtility
from wwdtm.validation import valid_int_id


class PanelistScores:
    """This class contains functions used to retrieve panelist scores
    from a copy of the Wait Wait Stats database.

    :param connect_dict: Dictionary containing database connection
        settings as required by mysql.connector.connect
    :param database_connection: mysql.connector.connect database
        connection
    """

    def __init__(
        self,
        connect_dict: Optional[Dict[str, Any]] = None,
        database_connection: Optional[connect] = None,
    ):
        """Class initialization method."""
        if connect_dict:
            self.connect_dict = connect_dict
            self.database_connection = connect(**connect_dict)
        elif database_connection:
            if not database_connection.is_connected():
                database_connection.reconnect()

            self.database_connection = database_connection

        self.utility = PanelistUtility(database_connection=self.database_connection)

    @lru_cache(typed=True)
    def retrieve_scores_by_id(self, panelist_id: int) -> List[int]:
        """Returns a list of panelist scores for appearances for the
        requested panelist ID.

        :param panelist_id: Panelist ID
        :return: List containing panelist scores. If panelist scores
            could not be retrieved, an empty list is returned.
        """
        if not valid_int_id(panelist_id):
            return []

        query = """
            SELECT pm.panelistscore AS score
            FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE panelistid = %s
            AND s.bestof = 0 and s.repeatshowid IS NULL
            ORDER BY s.showdate ASC;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query, (panelist_id,))
        result = cursor.fetchall()
        cursor.close()

        if not result:
            return []

        scores = []
        for appearance in result:
            if appearance.score:
                scores.append(appearance.score)

        return scores

    @lru_cache(typed=True)
    def retrieve_scores_by_slug(self, panelist_slug: str) -> List[int]:
        """Returns a list of panelist scores for appearances for the
        requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: List containing panelist scores. If panelist scores
            could not be retrieved, an empty list is returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return []

        return self.retrieve_scores_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_scores_grouped_list_by_id(
        self, panelist_id: int
    ) -> Dict[str, List[int]]:
        """Returns a panelist's score grouping for the requested
        panelist ID.

        :param panelist_id: Panelist ID
        :return: Dictionary containing two lists, one containing scores
            and one containing counts of those scores. If panelist
            scores could not be retrieved, an empty dictionary is
            returned.
        """
        if not valid_int_id(panelist_id):
            return {}

        query = """
            SELECT MIN(pm.panelistscore) AS min,
            MAX(pm.panelistscore) AS max
            FROM ww_showpnlmap pm
            LIMIT 1;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query)
        result = cursor.fetchone()

        if not result:
            return {}

        min_score = result.min
        max_score = result.max

        scores = {}
        for score in range(min_score, max_score + 1):
            scores[score] = 0

        query = """
            SELECT pm.panelistscore AS score,
            COUNT(pm.panelistscore) AS score_count
            FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s
            AND s.bestof = 0 AND s.repeatshowid IS NULL
            AND pm.panelistscore IS NOT NULL
            GROUP BY pm.panelistscore
            ORDER BY pm.panelistscore ASC;
            """
        cursor.execute(query, (panelist_id,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return {}

        for row in results:
            scores[row.score] = row.score_count

        return {
            "score": list(scores.keys()),
            "count": list(scores.values()),
        }

    @lru_cache(typed=True)
    def retrieve_scores_grouped_list_by_slug(
        self, panelist_slug: str
    ) -> Dict[str, List[int]]:
        """Returns a panelist's score grouping for the requested
        panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: Dictionary containing two lists, one containing scores
            and one containing counts of those scores. If panelist
            scores could not be retrieved, an empty dictionary is
            returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return {}

        return self.retrieve_scores_grouped_list_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_scores_grouped_ordered_pair_by_id(
        self, panelist_id: int
    ) -> List[Tuple[int, int]]:
        """Returns a list of tuples containing a score and the
        corresponding number of instances a panelist has scored that amount
        for the requested panelist ID.

        :param panelist_id: Panelist ID
        :return: List of tuples containing scores and score counts. If
            panelist scores could not be retrieved, an empty list is
            returned.
        """
        if not valid_int_id(panelist_id):
            return []

        query = """
            SELECT MIN(pm.panelistscore) AS min,
            MAX(pm.panelistscore) AS max
            FROM ww_showpnlmap pm;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query)
        result = cursor.fetchone()

        if not result:
            return []

        min_score = result.min
        max_score = result.max

        scores = {}
        for score in range(min_score, max_score + 1):
            scores[score] = 0

        query = """
            SELECT pm.panelistscore AS score,
            COUNT(pm.panelistscore) AS score_count
            FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s
            AND s.bestof = 0 AND s.repeatshowid IS NULL
            AND pm.panelistscore IS NOT NULL
            GROUP BY pm.panelistscore
            ORDER BY pm.panelistscore ASC;
            """
        cursor.execute(query, (panelist_id,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        for row in results:
            scores[row.score] = row.score_count

        return list(scores.items())

    @lru_cache(typed=True)
    def retrieve_scores_grouped_ordered_pair_by_slug(
        self,
        panelist_slug: str,
    ) -> List[Tuple[int, int]]:
        """Returns a list of tuples containing a score and the
        corresponding number of instances a panelist has scored that amount
        for the requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: List of tuples containing scores and score counts. If
            panelist scores could not be retrieved, an empty list is
            returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return []

        return self.retrieve_scores_grouped_ordered_pair_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_scores_list_by_id(
        self,
        panelist_id: int,
    ) -> Dict[str, List]:
        """Returns a dictionary containing two lists, one with show
        dates and one with corresponding scores for the requested
        panelist ID.

        :param panelist_id: Panelist ID
        :return: Dictionary containing a list show dates and a list
            of scores. If panelist scores could not be retrieved, an
            empty dictionary is returned.
        """
        if not valid_int_id(panelist_id):
            return {}

        query = """
            SELECT s.showdate AS date, pm.panelistscore AS score
            FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s
            AND s.bestof = 0 AND s.repeatshowid IS NULL
            AND pm.panelistscore IS NOT NULL
            ORDER BY s.showdate ASC;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query, (panelist_id,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return {}

        show_list = []
        score_list = []
        for shows in results:
            show_list.append(shows.date.isoformat())
            score_list.append(shows.score)

        return {
            "shows": show_list,
            "scores": score_list,
        }

    @lru_cache(typed=True)
    def retrieve_scores_list_by_slug(
        self,
        panelist_slug: str,
    ) -> Dict[str, List]:
        """Returns a dictionary containing two lists, one with show
        dates and one with corresponding scores for the requested
        panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: Dictionary containing a list show dates and a list
            of scores. If panelist scores could not be retrieved, an
            empty dictionary is returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return {}

        return self.retrieve_scores_list_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_scores_ordered_pair_by_id(
        self, panelist_id: int
    ) -> List[Tuple[str, int]]:
        """Returns a list of tuples containing a show date and the
        corresponding score for the requested panelist ID.

        :param panelist_id: Panelist ID
        :return: List of tuples containing show dates and scores. If
            panelist scores could not be retrieved, an empty list is
            returned.
        """
        if not valid_int_id(panelist_id):
            return []

        query = """
            SELECT s.showdate AS date, pm.panelistscore AS score
            FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s
            AND s.bestof = 0 AND s.repeatshowid IS NULL
            AND pm.panelistscore IS NOT NULL
            ORDER BY s.showdate ASC;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query, (panelist_id,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return []

        scores = []
        for show in results:
            show_date = show.date.isoformat()
            score = show.score
            scores.append((show_date, score))

        return scores

    @lru_cache(typed=True)
    def retrieve_scores_ordered_pair_by_slug(
        self,
        panelist_slug: str,
    ) -> List[Tuple[str, int]]:
        """Returns a list of tuples containing a show date and the
        corresponding score for the requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: List of tuples containing show dates and scores. If
            panelist scores could not be retrieved, an empty list is
            returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return []

        return self.retrieve_scores_ordered_pair_by_id(id_)



# -*- coding: utf-8 -*-
# vim: set noai syntax=python ts=4 sw=4:
#
# Copyright (c) 2018-2023 Linh Pham
# wwdtm is released under the terms of the Apache License 2.0
"""Wait Wait Don't Tell Me! Stats Panelist Statistics Retrieval Functions
"""
from decimal import Decimal
from functools import lru_cache
from typing import Any, Dict, Optional

from mysql.connector import connect, DatabaseError
import numpy
from wwdtm.panelist.scores import PanelistScores
from wwdtm.panelist.decimal_scores import PanelistDecimalScores
from wwdtm.panelist.utility import PanelistUtility
from wwdtm.validation import valid_int_id


class PanelistStatistics:
    """This class contains functions used to retrieve data from a copy
    of the Wait Wait Stats database and calculate statistics for
    panelists.

    :param connect_dict: Dictionary containing database connection
        settings as required by mysql.connector.connect
    :param database_connection: mysql.connector.connect database
        connection
    """

    def __init__(
        self,
        connect_dict: Optional[Dict[str, Any]] = None,
        database_connection: Optional[connect] = None,
    ):
        """Class initialization method."""
        if connect_dict:
            self.connect_dict = connect_dict
            self.database_connection = connect(**connect_dict)
        elif database_connection:
            if not database_connection.is_connected():
                database_connection.reconnect()

            self.database_connection = database_connection

        try:
            query = "SHOW COLUMNS FROM ww_showpnlmap WHERE Field = 'panelistlrndstart_decimal';"
            cursor = self.database_connection.cursor()
            cursor.execute(query)
            start_decimal = cursor.fetchone()

            query = (
                "SHOW COLUMNS FROM ww_showpnlmap WHERE Field = 'panelistscore_decimal';"
            )
            cursor = self.database_connection.cursor()
            cursor.execute(query)
            score_decimal = cursor.fetchone()
            cursor.close()

            if start_decimal and score_decimal:
                self.has_decimal_columns: bool = True
            else:
                self.has_decimal_columns: bool = False
        except DatabaseError:
            self.has_decimal_columns: bool = False

        self.scores = PanelistScores(database_connection=self.database_connection)
        self.scores_decimal = PanelistDecimalScores(
            database_connection=self.database_connection
        )
        self.utility = PanelistUtility(database_connection=self.database_connection)

    @lru_cache(typed=True)
    def retrieve_bluffs_by_id(self, panelist_id: int) -> Dict[str, int]:
        """Returns a dictionary containing the number of chosen Bluffs
        and correct Bluffs for the requested panelist ID.

        :param panelist_id: Panelist ID
        :return: Dictionary containing panelist Bluff counts. If
            panelist Bluff counts could not be returned, an empty
            dictionary will be returned.
        """
        query = """
            SELECT (
            SELECT COUNT(blm.chosenbluffpnlid) FROM ww_showbluffmap blm
            JOIN ww_shows s ON s.showid = blm.showid
            WHERE s.repeatshowid IS NULL AND blm.chosenbluffpnlid = %s
            ) AS chosen, (
            SELECT COUNT(blm.correctbluffpnlid) FROM ww_showbluffmap blm
            JOIN ww_shows s ON s.showid = blm.showid
            WHERE s.repeatshowid IS NULL AND blm.correctbluffpnlid = %s
            ) AS correct;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(
            query,
            (
                panelist_id,
                panelist_id,
            ),
        )
        result = cursor.fetchone()
        cursor.close()

        if not result:
            return {}

        return {
            "chosen": result.chosen,
            "correct": result.correct,
        }

    @lru_cache(typed=True)
    def retrieve_bluffs_by_slug(self, panelist_slug: str) -> Dict[str, int]:
        """Returns a dictionary containing the number of chosen Bluffs
        and correct Bluffs for the requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: Dictionary containing panelist Bluff counts. If
            panelist Bluff counts could not be returned, an empty
            dictionary will be returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return {}

        return self.retrieve_bluffs_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_rank_info_by_id(self, panelist_id: int) -> Dict[str, int]:
        """Returns a dictionary with ranking information for the
        requested panelist ID.

        :param panelist_id: Panelist ID
        :return: Dictionary containing panelist ranking information. If
            panelist ranking information could not be returned, an empty
            dictionary will be returned.
        """
        if not valid_int_id(panelist_id):
            return {}

        query = """
            SELECT (
            SELECT COUNT(pm.showpnlrank) FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s AND pm.showpnlrank = '1' AND
            s.bestof = 0 and s.repeatshowid IS NULL) as 'first', (
            SELECT COUNT(pm.showpnlrank) FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s AND pm.showpnlrank = '1t' AND
            s.bestof = 0 and s.repeatshowid IS NULL) as 'first_tied', (
            SELECT COUNT(pm.showpnlrank) FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s AND pm.showpnlrank = '2' AND
            s.bestof = 0 and s.repeatshowid IS NULL) as 'second', (
            SELECT COUNT(pm.showpnlrank) FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s AND pm.showpnlrank = '2t' AND
            s.bestof = 0 and s.repeatshowid IS NULL) as 'second_tied', (
            SELECT COUNT(pm.showpnlrank) FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s AND pm.showpnlrank = '3' AND
            s.bestof = 0 and s.repeatshowid IS NULL
            ) as 'third';
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(
            query,
            (
                panelist_id,
                panelist_id,
                panelist_id,
                panelist_id,
                panelist_id,
            ),
        )
        result = cursor.fetchone()
        cursor.close()

        if not result:
            return {}

        return {
            "first": result.first,
            "first_tied": result.first_tied,
            "second": result.second,
            "second_tied": result.second_tied,
            "third": result.third,
        }

    @lru_cache(typed=True)
    def retrieve_rank_info_by_slug(self, panelist_slug: str) -> Dict[str, int]:
        """Returns a dictionary with ranking information for the
        requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: Dictionary containing panelist ranking information. If
            panelist ranking information could not be returned, an empty
            dictionary will be returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return {}

        return self.retrieve_rank_info_by_id(id_)

    @lru_cache(typed=True)
    def retrieve_statistics_by_id(
        self, panelist_id: int, include_decimal_scores: bool = False
    ) -> Dict[str, Any]:
        """Returns a dictionary containing panelist statistics, ranking
        data, and scoring data for the requested panelist ID.

        :param panelist_id: Panelist ID
        :param include_decimal_scores: Flag set to include statistics
            based on decimal scores along with statistics based on
            integer scores
        :return: Dictionary containing panelist statistics. If panelist
            statistics could not be returned, an empty dictionary will
            be returned.
        """
        if not valid_int_id(panelist_id):
            return {}

        score_data = self.scores.retrieve_scores_by_id(panelist_id)
        ranks = self.retrieve_rank_info_by_id(panelist_id)
        if not score_data or not ranks:
            return {}

        if self.has_decimal_columns and include_decimal_scores:
            score_data_decimal = self.scores_decimal.retrieve_scores_by_id(panelist_id)
            if not score_data_decimal:
                return {}

        appearance_count = len(score_data)
        scoring = {
            "minimum": int(numpy.amin(score_data)),
            "maximum": int(numpy.amax(score_data)),
            "mean": round(numpy.mean(score_data), 5),
            "median": int(numpy.median(score_data)),
            "standard_deviation": round(numpy.std(score_data), 5),
            "total": int(numpy.sum(score_data)),
        }

        if include_decimal_scores:
            scoring_decimal = {
                "minimum": Decimal(numpy.amin(score_data_decimal)),
                "maximum": Decimal(numpy.amax(score_data_decimal)),
                "mean": round(Decimal(numpy.mean(score_data_decimal)), 5),
                "median": Decimal(numpy.median(score_data_decimal)),
                "standard_deviation": round(Decimal(numpy.std(score_data_decimal)), 5),
                "total": Decimal(numpy.sum(score_data_decimal)),
            }

        ranks_first = round(100 * (ranks["first"] / appearance_count), 5)
        ranks_first_tied = round(100 * (ranks["first_tied"] / appearance_count), 5)
        ranks_second = round(100 * (ranks["second"] / appearance_count), 5)
        ranks_second_tied = round(100 * (ranks["second_tied"] / appearance_count), 5)
        ranks_third = round(100 * (ranks["third"] / appearance_count), 5)

        ranks_percentage = {
            "first": ranks_first,
            "first_tied": ranks_first_tied,
            "second": ranks_second,
            "second_tied": ranks_second_tied,
            "third": ranks_third,
        }

        ranking = {
            "rank": ranks,
            "percentage": ranks_percentage,
        }

        if not include_decimal_scores:
            return {
                "scoring": scoring,
                "ranking": ranking,
            }
        else:
            return {
                "scoring": scoring,
                "scoring_decimal": scoring_decimal,
                "ranking": ranking,
            }

    @lru_cache(typed=True)
    def retrieve_statistics_by_slug(
        self, panelist_slug: str, include_decimal_scores: bool = False
    ) -> Dict[str, Any]:
        """Returns a dictionary containing panelist statistics, ranking
        data, and scoring data for the requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :param include_decimal_scores: Flag set to include statistics
            based on decimal scores along with statistics based on
            integer scores
        :return: Dictionary containing panelist statistics. If panelist
            statistics could not be returned, an empty dictionary will
            be returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return {}

        return self.retrieve_statistics_by_id(
            id_, include_decimal_scores=include_decimal_scores
        )



# -*- coding: utf-8 -*-
# vim: set noai syntax=python ts=4 sw=4:
#
# Copyright (c) 2018-2023 Linh Pham
# wwdtm is released under the terms of the Apache License 2.0
"""Wait Wait Don't Tell Me! Stats Panelist Appearance Retrieval Functions
"""
from functools import lru_cache
from typing import Any, Dict, Optional

from mysql.connector import connect, DatabaseError
from wwdtm.panelist.utility import PanelistUtility
from wwdtm.validation import valid_int_id


class PanelistAppearances:
    """This class contains functions that retrieve panelist appearance
    information from a copy of the Wait Wait Stats database.

    :param connect_dict: Dictionary containing database connection
        settings as required by mysql.connector.connect
    :param database_connection: mysql.connector.connect database
        connection
    """

    def __init__(
        self,
        connect_dict: Optional[Dict[str, Any]] = None,
        database_connection: Optional[connect] = None,
    ):
        """Class initialization method."""
        if connect_dict:
            self.connect_dict = connect_dict
            self.database_connection = connect(**connect_dict)
        elif database_connection:
            if not database_connection.is_connected():
                database_connection.reconnect()

            self.database_connection = database_connection

        try:
            query = "SHOW COLUMNS FROM ww_showpnlmap WHERE Field = 'panelistlrndstart_decimal';"
            cursor = self.database_connection.cursor()
            cursor.execute(query)
            start_decimal = cursor.fetchone()

            query = (
                "SHOW COLUMNS FROM ww_showpnlmap WHERE Field = 'panelistscore_decimal';"
            )
            cursor = self.database_connection.cursor()
            cursor.execute(query)
            score_decimal = cursor.fetchone()
            cursor.close()

            if start_decimal and score_decimal:
                self.has_decimal_columns: bool = True
            else:
                self.has_decimal_columns: bool = False
        except DatabaseError:
            self.has_decimal_columns: bool = False

        self.utility = PanelistUtility(database_connection=self.database_connection)

    @lru_cache(typed=True)
    def retrieve_appearances_by_id(
        self, panelist_id: int, use_decimal_scores: bool = False
    ) -> Dict[str, Any]:
        """Returns a list of dictionary objects containing appearance
        information for the requested panelist ID.

        :param panelist_id: Panelist ID
        :param use_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: Dictionary containing appearance counts and list of
            appearances for a panelist. If panelist appearances could
            not be retrieved, an empty dictionary is returned.
        """
        if not valid_int_id(panelist_id):
            return {}

        query = """
            SELECT (
            SELECT COUNT(pm.showid) FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE s.bestof = 0 AND s.repeatshowid IS NULL AND
            pm.panelistid = %s ) AS regular_shows, (
            SELECT COUNT(pm.showid) FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE pm.panelistid = %s ) AS all_shows, (
            SELECT COUNT(pm.panelistid) FROM ww_showpnlmap pm
            JOIN ww_shows s ON pm.showid = s.showid
            WHERE pm.panelistid = %s AND s.bestof = 0 AND
            s.repeatshowid IS NULL
            AND pm.panelistscore IS NOT NULL )
            AS shows_with_scores;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(
            query,
            (
                panelist_id,
                panelist_id,
                panelist_id,
            ),
        )
        result = cursor.fetchone()

        if result:
            appearance_counts = {
                "regular_shows": result.regular_shows,
                "all_shows": result.all_shows,
                "shows_with_scores": result.shows_with_scores,
            }
        else:
            appearance_counts = {
                "regular_shows": 0,
                "all_shows": 0,
                "shows_with_scores": 0,
            }

        query = """
            SELECT MIN(s.showid) AS first_id, MIN(s.showdate) AS first,
            MAX(s.showid) AS most_recent_id, MAX(s.showdate) AS most_recent
            FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            WHERE s.bestof = 0 AND s.repeatshowid IS NULL
            AND pm.panelistid = %s
            ORDER BY s.showdate ASC;
            """
        cursor.execute(query, (panelist_id,))
        result = cursor.fetchone()

        if result and result.first_id:
            first = {
                "show_id": result.first_id,
                "show_date": result.first.isoformat(),
            }

            most_recent = {
                "show_id": result.most_recent_id,
                "show_date": result.most_recent.isoformat(),
            }

            milestones = {
                "first": first,
                "most_recent": most_recent,
            }

            appearance_info = {
                "milestones": milestones,
            }
        else:
            appearance_info = {
                "milestones": None,
            }

        if use_decimal_scores and self.has_decimal_columns:
            query = """
                SELECT pm.showid AS show_id, s.showdate AS date,
                s.bestof AS best_of, s.repeatshowid AS repeat_show_id,
                pm.panelistlrndstart AS start,
                pm.panelistlrndstart_decimal AS start_decimal,
                pm.panelistlrndcorrect AS correct,
                pm.panelistlrndcorrect_decimal AS correct_decimal,
                pm.panelistscore AS score,
                pm.panelistscore_decimal AS score_decimal,
                pm.showpnlrank AS pnl_rank FROM ww_showpnlmap pm
                JOIN ww_panelists p ON p.panelistid = pm.panelistid
                JOIN ww_shows s ON s.showid = pm.showid
                WHERE pm.panelistid = %s
                ORDER BY s.showdate ASC;
                """
        else:
            query = """
                SELECT pm.showid AS show_id, s.showdate AS date,
                s.bestof AS best_of, s.repeatshowid AS repeat_show_id,
                pm.panelistlrndstart AS start,
                pm.panelistlrndcorrect AS correct,
                pm.panelistscore AS score,
                pm.showpnlrank AS pnl_rank FROM ww_showpnlmap pm
                JOIN ww_panelists p ON p.panelistid = pm.panelistid
                JOIN ww_shows s ON s.showid = pm.showid
                WHERE pm.panelistid = %s
                ORDER BY s.showdate ASC;
                """

        cursor.execute(query, (panelist_id,))
        results = cursor.fetchall()
        cursor.close()

        if result:
            appearances = []
            for appearance in results:
                info = {
                    "show_id": appearance.show_id,
                    "date": appearance.date.isoformat(),
                    "best_of": bool(appearance.best_of),
                    "repeat_show": bool(appearance.repeat_show_id),
                    "lightning_round_start": appearance.start,
                    "lightning_round_start_decimal": appearance.start_decimal
                    if "start_decimal" in appearance._fields
                    else None,
                    "lightning_round_correct": appearance.correct,
                    "lightning_round_correct_decimal": appearance.correct_decimal
                    if "correct_decimal" in appearance._fields
                    else None,
                    "score": appearance.score,
                    "score_decimal": appearance.score_decimal
                    if "score_decimal" in appearance._fields
                    else None,
                    "rank": appearance.pnl_rank
                    if appearance.pnl_rank is not None
                    else None,
                }
                appearances.append(info)

            appearance_info["count"] = appearance_counts
            appearance_info["shows"] = appearances
        else:
            appearance_info["count"] = appearance_counts
            appearance_info["shows"] = []

        return appearance_info

    @lru_cache(typed=True)
    def retrieve_appearances_by_slug(
        self, panelist_slug: str, use_decimal_scores: bool = False
    ) -> Dict[str, Any]:
        """Returns a list of dictionary objects containing appearance
        information for the requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :param use_decimal_scores: Flag set to include panelist decimal
            scores, if available
        :return: Dictionary containing appearance counts and list of
            appearances for a panelist. If panelist appearances could
            not be retrieved, an empty dictionary is returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return {}

        return self.retrieve_appearances_by_id(
            id_, use_decimal_scores=use_decimal_scores
        )

    @lru_cache(typed=True)
    def retrieve_yearly_appearances_by_id(self, panelist_id: int) -> Dict[int, int]:
        """Returns a dictionary containing panelist appearances broken
        down by year, for the requested panelist ID.

        :param panelist_id: Panelist ID
        :return: Dictionary containing scoring breakdown by year. If
            panelist appearances could not be retrieved, an empty
            dictionary is returned.
        """
        if not valid_int_id(panelist_id):
            return {}

        query = """
            SELECT DISTINCT YEAR(s.showdate) AS year
            FROM ww_shows s
            ORDER BY YEAR(s.showdate) ASC;
            """
        cursor = self.database_connection.cursor(named_tuple=True)
        cursor.execute(query)
        results = cursor.fetchall()

        if not results:
            return {}

        years = {}
        for row in results:
            years[row.year] = 0

        query = """
            SELECT YEAR(s.showdate) AS year,
            COUNT(p.panelist) AS count
            FROM ww_showpnlmap pm
            JOIN ww_shows s ON s.showid = pm.showid
            JOIN ww_panelists p ON p.panelistid = pm.panelistid
            WHERE pm.panelistid = %s AND s.bestof = 0
            AND s.repeatshowid IS NULL
            GROUP BY p.panelist, YEAR(s.showdate)
            ORDER BY p.panelist ASC, YEAR(s.showdate) ASC;
            """
        cursor.execute(query, (panelist_id,))
        results = cursor.fetchall()
        cursor.close()

        if not results:
            return {}

        for row in results:
            years[row.year] = row.count

        return years

    @lru_cache(typed=True)
    def retrieve_yearly_appearances_by_slug(self, panelist_slug: str) -> Dict[int, int]:
        """Returns a dictionary containing panelist appearances broken
        down by year, for the requested panelist slug string.

        :param panelist_slug: Panelist slug string
        :return: Dictionary containing scoring breakdown by year. If
            panelist appearances could not be retrieved, an empty
            dictionary is returned.
        """
        id_ = self.utility.convert_slug_to_id(panelist_slug)
        if not id_:
            return {}

        return self.retrieve_yearly_appearances_by_id(id_)



# -*- coding: utf-8 -*-
# vim: set noai syntax=python ts=4 sw=4:
#
# Copyright (c) 2018-2023 Linh Pham
# wwdtm is released under the terms of the Apache License 2.0
"""Testing for object: :py:class:`wwdtm.panelist.PanelistScores`
"""
import json
from typing import Any, Dict

import pytest
from wwdtm.panelist import PanelistScores


@pytest.mark.skip
def get_connect_dict() -> Dict[str, Any]:
    """Read in database connection settings and return values as a
    dictionary.

    :return: A dictionary containing database connection settings
        for use by mysql.connector
    """
    with open("config.json", "r", encoding="utf-8") as config_file:
        config_dict = json.load(config_file)
        if "database" in config_dict:
            return config_dict["database"]


@pytest.mark.parametrize("panelist_id", [14])
def test_panelist_scores_retrieve_scores_by_id(panelist_id: int):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_by_id`

    :param panelist_id: Panelist ID to test retrieving panelist
        information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_by_id(panelist_id)

    assert scoring, f"Scoring data not returned for ID {panelist_id}"


@pytest.mark.parametrize("panelist_slug", ["luke-burbank"])
def test_panelist_scores_retrieve_scores_by_slug(panelist_slug: str):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_by_slug`

    :param panelist_slug: Panelist slug string to test retrieving
        panelist information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_by_slug(panelist_slug)

    assert scoring, f"Scoring data not returned for slug {panelist_slug}"


@pytest.mark.parametrize("panelist_id", [14])
def test_panelist_scores_retrieve_scores_grouped_list_by_id(panelist_id: int):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_grouped_list_by_id`

    :param panelist_id: Panelist ID to test retrieving panelist
        information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_grouped_list_by_id(panelist_id)

    assert "score" in scoring, f"'score' was not returned for ID {panelist_id}"
    assert "count" in scoring, f"'count' was not returned for ID {panelist_id}"


@pytest.mark.parametrize("panelist_slug", ["luke-burbank"])
def test_panelist_scores_retrieve_scores_grouped_list_by_slug(panelist_slug: str):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_grouped_list_by_slug`

    :param panelist_slug: Panelist slug string to test retrieving
        panelist information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_grouped_list_by_slug(panelist_slug)

    assert "score" in scoring, f"'score' was not returned for slug {panelist_slug}"
    assert "count" in scoring, f"'count' was not returned for slug {panelist_slug}"


@pytest.mark.parametrize("panelist_id", [14])
def test_panelist_scores_retrieve_scores_grouped_ordered_pair_by_id(panelist_id: int):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_grouped_ordered_pair_by_id`

    :param panelist_id: Panelist ID to test retrieving panelist
        information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_grouped_ordered_pair_by_id(panelist_id)

    assert scoring, f"Scoring data not returned for ID {panelist_id}"
    assert isinstance(scoring[0], tuple), "First list item is not a tuple"


@pytest.mark.parametrize("panelist_slug", ["luke-burbank"])
def test_panelist_scores_retrieve_scores_grouped_ordered_pair_by_slug(
    panelist_slug: str,
):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_grouped_ordered_pair_by_slug`

    :param panelist_slug: Panelist slug string to test retrieving
        panelist information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_grouped_ordered_pair_by_slug(panelist_slug)

    assert scoring, f"Scoring data not returned for slug {panelist_slug}"
    assert isinstance(scoring[0], tuple), "First list item is not a tuple"


@pytest.mark.parametrize("panelist_id", [14])
def test_panelist_scores_retrieve_scores_list_by_id(panelist_id: int):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_list_by_id`

    :param panelist_id: Panelist ID to test retrieving panelist information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_list_by_id(panelist_id)

    assert "shows" in scoring, f"'shows' was not returned for ID {panelist_id}"
    assert "scores" in scoring, f"'scores' was not returned for ID {panelist_id}"


@pytest.mark.parametrize("panelist_slug", ["luke-burbank"])
def test_panelist_scores_retrieve_scores_list_by_slug(panelist_slug: str):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_list_by_slug`

    :param panelist_slug: Panelist slug string to test retrieving
        panelist information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_list_by_slug(panelist_slug)

    assert "shows" in scoring, f"'shows' was not returned for slug {panelist_slug}"
    assert "scores" in scoring, f"'scores' was not returned for slug {panelist_slug}"


@pytest.mark.parametrize("panelist_id", [14])
def test_panelist_scores_retrieve_scores_ordered_pair_by_id(panelist_id: int):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_ordered_pair_by_id`

    :param panelist_id: Panelist ID to test retrieving panelist
        information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_ordered_pair_by_id(panelist_id)

    assert scoring, f"Scoring data not returned for ID {panelist_id}"


@pytest.mark.parametrize("panelist_slug", ["luke-burbank"])
def test_panelist_scores_retrieve_scores_ordered_pair_by_slug(panelist_slug: str):
    """Testing for :py:meth:`wwdtm.panelist.PanelistScores.retrieve_scores_ordered_pair_by_slug`

    :param panelist_slug: Panelist slug string to test retrieving panelist
        information
    """
    scores = PanelistScores(connect_dict=get_connect_dict())
    scoring = scores.retrieve_scores_ordered_pair_by_slug(panelist_slug)

    assert scoring, f"Scoring data not returned for slug {panelist_slug}"
